package limax.switcher;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.nio.file.FileVisitResult;
import java.nio.file.FileVisitor;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.attribute.BasicFileAttributes;

import limax.codec.MarshalException;
import limax.codec.Octets;
import limax.codec.OctetsStream;
import limax.pkix.TrustManager;
import limax.util.ZipUtils;

public class LmkConfigData {
	private final Octets zipTrustsPath;
	private final String revocationCheckerOptions;
	private final boolean validateDate;

	public LmkConfigData(Path trustsPath, String revocationCheckerOptions, boolean validateDate) throws IOException {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ZipUtils.zip(trustsPath, baos);
		this.zipTrustsPath = Octets.wrap(baos.toByteArray());
		this.revocationCheckerOptions = revocationCheckerOptions;
		this.validateDate = validateDate;
	}

	public LmkConfigData(Octets binary) throws MarshalException {
		OctetsStream os = new OctetsStream(binary);
		this.zipTrustsPath = os.unmarshal_Octets();
		this.revocationCheckerOptions = os.unmarshal_String();
		this.validateDate = os.unmarshal_boolean();
	}

	public Octets encode() {
		return new OctetsStream().marshal(zipTrustsPath).marshal(revocationCheckerOptions).marshal(validateDate);
	}

	public LmkMasquerade createLmkMasquerade() throws IOException {
		TrustManager trustManager = new TrustManager();
		trustManager.setRecovationCheckerOptions(revocationCheckerOptions);
		Path trustsPath = Files.createTempDirectory("trustsPath");
		ZipUtils.unzip(new ByteArrayInputStream(zipTrustsPath.getBytes()), trustsPath);
		trustManager.addTrust(trustsPath);
		Files.walkFileTree(trustsPath, new FileVisitor<Path>() {
			@Override
			public FileVisitResult preVisitDirectory(Path dir, BasicFileAttributes attrs) throws IOException {
				return FileVisitResult.CONTINUE;
			}

			@Override
			public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
				Files.deleteIfExists(file);
				return FileVisitResult.CONTINUE;
			}

			@Override
			public FileVisitResult visitFileFailed(Path file, IOException exc) throws IOException {
				return FileVisitResult.CONTINUE;
			}

			@Override
			public FileVisitResult postVisitDirectory(Path dir, IOException exc) throws IOException {
				Files.deleteIfExists(dir);
				return FileVisitResult.CONTINUE;
			}
		});
		Files.deleteIfExists(trustsPath);
		return new LmkMasquerade(trustManager, validateDate);
	}
}
