package limax.switcher;

import java.util.Queue;
import java.util.concurrent.LinkedBlockingQueue;

import limax.codec.Octets;
import limax.net.Engine;
import limax.net.Transport;
import limax.switcher.switcherauany.Exchange;
import limax.switcherauany.AuanyAuthArg;

class LmkManager {
	private final static LmkManager instance = new LmkManager();

	private LmkManager() {
	}

	static LmkManager getInstance() {
		return instance;
	}

	private final Queue<Exchange> queue = new LinkedBlockingQueue<>();
	private LmkMasquerade lmkMasquerade;

	void setup(LmkMasquerade lmkMasquerade) {
		this.lmkMasquerade = lmkMasquerade;
	}

	boolean inspect(AuanyAuthArg auth, Octets passphrase) {
		return lmkMasquerade == null || !auth.platflag.startsWith("lmk")
				|| lmkMasquerade.masquerade(auth.username, auth.token, passphrase, (username, token, notAfter) -> {
					auth.username = username;
					auth.token = token;
				});
	}

	private void flush(Transport transport) {
		for (Exchange upload; (upload = queue.poll()) != null;) {
			try {
				upload.send(transport);
			} catch (Exception e) {
				queue.offer(upload);
			}
		}
	}

	void auanyOnline(Transport transport) {
		flush(transport);
	}

	void ack(String uid) {
		queue.offer(new Exchange(Exchange.UPLOAD_LMKDATA, new Octets()));
		Engine.getApplicationExecutor().execute(() -> {
			Transport transport = AuanyClientListener.getInstance().getTransport();
			if (transport != null)
				flush(transport);
		});
	}
}
